from dkist_fits_specifications.spec122 import (
    define_122_schema_expansions,
    load_processed_spec122,
    load_raw_spec122,
    load_spec122,
)
from dkist_fits_specifications.utils.frozendict import frozendict


def test_load_raw_122():
    """
    Given: The raw SPEC-122 yamls
    When: Loading the raw yamls
    Then: The correct return types and structure are returned
    """
    spec = load_raw_spec122()
    assert isinstance(spec, frozendict)
    assert isinstance(spec["fits"], tuple)
    header, spec = spec["fits"]
    assert isinstance(header, frozendict)
    assert isinstance(spec["NAXIS"], frozendict)


def test_load_122():
    """
    Given: The raw SPEC-122 yamls
    When: Loading pre-processed spec sections
    Then: The section headers have been applied but expansion has NOT been done
    """
    spec = load_spec122()
    assert isinstance(spec, frozendict)
    assert isinstance(spec["fits"], frozendict)
    assert isinstance(spec["fits"]["NAXIS"], frozendict)

    # Make sure the header was applied
    assert spec["vbi"]["VBI__001"]["instrument_required"] == "VBI"
    assert spec["vbi"]["VBI__001"]["required"] is False

    # Make sure no expansion happened
    assert "NAXIS3" not in spec["fits"]
    assert "NAXIS<n>" in spec["fits"]
    assert "ZNAXIS<n>" in spec["compression"]
    assert "PC<i>_<j>" in spec["telescope"]


def test_load_processed_122():
    """
    Given: The raw SPEC-122 yamls and a header with an INSTRUME value
    When: Loading the fully-processed SPEC-122 schema
    Then: Expansion and conditional requiredness are correctly applied
    """
    spec = load_processed_spec122(INSTRUME="VBI")

    assert spec["vbi"]["VBI__001"]["required"] is True
    assert spec["visp"]["VISP_001"]["required"] is False

    # Make sure expansion worked. One key for each expansion index for each section *should* be fine
    # We use the max value for each index.
    assert "NAXIS3" in spec["fits"]
    assert "ZNAXIS3" in spec["compression"]
    assert "CRPIX3" in spec["telescope"]
    assert "PC3_3" in spec["telescope"]


def test_define_122_schema_expansion_duplication():
    """
    Given: the list of requested spec 122 expansions
    When: checking the indices for each expansion
    Then: None of them match (all expansions are unique)
    """
    expansions = define_122_schema_expansions()
    expansion_indices = [e.index for e in expansions]
    assert len(expansion_indices) == len(set(expansion_indices))



"""
Functions and schemas relating to specification 122, the level 0 FITS files.

The 122 schemas have a singular variant, they are all NAXIS=3.
However, the yaml files still are written with the indices corresponding to NAXIS (``n, i, j``).
These indices are expanded by the loading function so the returned schema always has ``NAXIS = 3``.

The yamls are written with the indices, to reduce duplication and to make it
easier for 214 to use the raw (unprocessed) schemas.
"""
import copy
from typing import Any, Optional
from pathlib import Path
from functools import cache

import yamale  # type: ignore

from dkist_fits_specifications import validation_schema
from dkist_fits_specifications.utils import (
    frozendict,
    load_raw_spec,
    raw_schema_type_hint,
    schema_type_hint,
)

__all__ = ["load_raw_spec122", "load_spec122"]

from dkist_fits_specifications.utils.spec_processors.expansion import (
    ExpansionIndex,
    expand_schema,
)
from dkist_fits_specifications.utils.spec_processors.instrument_requiredness import (
    InstrumentRequired,
)
from dkist_fits_specifications.utils.spec_processors.requiredness_base import (
    ConditionalRequirement,
    update_schema_requiredness,
)

base_path = Path(__file__).parent / "schemas"


def preprocess_schema(schema: schema_type_hint) -> schema_type_hint:
    """
    Convert the loaded raw schemas to the 122 schema.

    This simply means applying each section's header to its keys.

    Parameters
    ----------
    raw_schema
        The loaded version of a single yaml file.

    Returns
    -------
    schema
        The body of a schema, updated as needed from the yaml files.
    """
    header, raw_schemas = schema
    header = dict(copy.deepcopy(header)["spec122"])
    header.pop("section")

    # Thaw the schema
    schema = dict(raw_schemas)
    for key, key_schema in schema.items():
        updated_schema = {key: {**header, **key_schema}}
        # Rather than put expected in all the files, default it to required
        updated_schema[key]["expected"] = key_schema.get("expected", key_schema["required"])
        updated_schema[key] = frozendict(updated_schema[key])
        schema.update(updated_schema)
    return frozendict(schema)


# No cache here as load_raw_spec is cached
def load_raw_spec122(glob: Optional[str] = None) -> frozendict[str, raw_schema_type_hint]:
    """
    Load the raw 122 schemas from the yaml files.

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.

    Returns
    -------
    raw_schemas
        The schemas as loaded from the yaml files.
    """
    return load_raw_spec(base_path, glob)


@cache
def load_spec122(glob: Optional[str] = None) -> frozendict[str, schema_type_hint]:
    """
    Return the loaded schemas for DKIST Specification 122

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.
    """

    raw_schemas = load_raw_spec122(glob=glob)
    schemas = {}
    for schema_name, raw_schema in raw_schemas.items():
        # 122 only uses the second document
        schema = preprocess_schema(raw_schema)
        yamale.validate(validation_schema, [(schema, None)])
        schemas[schema_name] = schema

    return frozendict(schemas)


def load_processed_spec122(
    glob: Optional[str] = None, **header: dict[str, Any]
) -> dict[str, schema_type_hint]:
    """
    Load the 122 schema, processed based on a FITS header.

    This function currently expands the schema and applies conditional requiredness

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.

    header
       A FITS header to be used to expand the schema as needed.

    Notes
    -----
    If you process the schema with a single header from a dataset, the processed
    schema should be valid for all the files in that dataset.
    """
    schemas = load_spec122(glob)
    processed_schemas = {}
    for section, schema in schemas.items():
        processed_schemas[section] = process_spec122_section(schema, **header)

    return processed_schemas


def process_spec122_section(schema: schema_type_hint, **header: dict[str, Any]) -> schema_type_hint:
    """
    Apply expansions and requiredness to a single section of the 122 schema.

    This is a separate function from `load_processed_spec122` to allow it to be used elsewhere in this package.
    """
    schema = update_122_schema_requiredness(schema, **header)
    schema = expand_122_schema(schema)

    return schema


def define_122_schema_expansions() -> list[ExpansionIndex]:
    # 122 always has 3 axes, but we encode the yamls independent of the number,
    # to match 214 which might have a variable number of axes.
    n_expansion = ExpansionIndex(index="n", size=1, values=range(1, 4))
    i_expansion = ExpansionIndex(index="i", size=1, values=range(1, 4))
    j_expansion = ExpansionIndex(index="j", size=1, values=range(1, 4))
    return [n_expansion, i_expansion, j_expansion]


def expand_122_schema(schema: schema_type_hint) -> dict[str, frozendict[str, Any]]:
    """
    Given a 122 header expand all known indices.
    """
    return expand_schema(schema=schema, expansions=define_122_schema_expansions())


def define_122_conditional_requiredness(**header: dict[str, Any]) -> list[ConditionalRequirement]:
    """
    Collate the list of conditional requirements to apply for spec 122
    """
    requirements_list = []

    instrument_requirement = InstrumentRequired(header)
    requirements_list.append(instrument_requirement)

    return requirements_list


def update_122_schema_requiredness(
    schema: schema_type_hint, **header: dict[str, Any]
) -> dict[str, frozendict[str, Any]]:
    """
    Given a 122 header, update the "required" spec field based on presence/values of spec fields that indicate conditional requiredness.
    """
    return update_schema_requiredness(
        schema=schema, requirements=define_122_conditional_requiredness(**header)
    )



from random import randint

import pytest
from astropy.io import fits

from dkist_fits_specifications.spec214 import load_processed_spec214
from dkist_fits_specifications.utils.formatter import reformat_spec214_header


@pytest.fixture
def header_with_all_sections() -> fits.Header:
    all_214_schemas = load_processed_spec214()
    header_dict = dict()
    for section in all_214_schemas.values():
        # Value doesn't matter at all, so just make it a random int
        header_dict.update({k: randint(1, 10) for k in section.keys()})

    # Manual intervention needed on these keys so header processors will work correctly
    del header_dict["END"]
    header_dict["DAAXES"] = 3
    header_dict["DEAXES"] = 2
    header_dict["DNAXIS"] = 5
    header_dict["NPROPOS"] = 10
    header_dict["NEXPERS"] = 11
    header_dict["NAXIS"] = 3
    header_dict["ZVAL1"] = 1
    header_dict["TFIELDS"] = 2
    header_dict["NSPECLNS"] = 13
    return fits.Header(header_dict)


def test_formatter(header_with_all_sections):
    """
    Given: A 214 header with all possible schema sections
    When: Formatting the header
    Then: The dang thing runs
    """
    output_header = reformat_spec214_header(header_with_all_sections)
    assert isinstance(output_header, fits.Header)



from abc import ABC, abstractmethod
from typing import Any

from dkist_fits_specifications.utils import frozendict, schema_type_hint

# Defined as a random string so we don't accidentally confuse ourselves in tests where we want to
# use a fake instrument name
_NOT_A_DKIST_INSTRUMENT = "67197693JustAFallbackInstrument062757b16"


class ConditionalRequirement(ABC):
    """
    A class definition for capturing a single type of conditional requiredness.

    These classes are instantiated with a header and then can be used to check the requiredness of any key in the schema
    """

    def __init__(self, header: dict[str:Any]):
        self.header = header

    @abstractmethod
    def check_requiredness(self, spec_fields: frozendict[str, Any]) -> bool:
        """Given the schema for a single key, check if this key needs to be required."""
        pass


def update_schema_requiredness(
    schema: schema_type_hint, requirements: list[ConditionalRequirement]
) -> dict[str, frozendict[str, Any]]:
    """
    Modify a schema's `required` values based on conditional requirements.

    If ANY of the conditional requiredness requirements are met then the key is set to "required."
    """
    updated_schema = dict()
    for fits_keyword_name, spec_fields in schema.items():
        thawed_fields = dict(spec_fields)

        if any([requirement.check_requiredness(spec_fields) for requirement in requirements]):
            thawed_fields["required"] = True

        updated_schema[fits_keyword_name] = frozendict(thawed_fields)

    return updated_schema



import logging
from typing import Any
from functools import partial

from dkist_fits_specifications.utils import frozendict
from dkist_fits_specifications.utils.spec_processors.requiredness_base import (
    _NOT_A_DKIST_INSTRUMENT,
    ConditionalRequirement,
)

logger = logging.getLogger(__name__)

# How this works:
# The top-level key is the SPEC-controlled instrument name. The value is a dict[str, Iterable and not str]
# This second-level dictionary defines a linking between header keys and values needed for a header to be considered
# "polarimetric". The logic is all([header[key] in second_level[key] for key in second_level.keys()]).
# For example, consider POLARIMETRIC_HEADER_REQUIREMENTS = {"instrument": {"key1": [1, 2], "key2": ["yes", "true"]}}
# In this case, a header would be considered polarimetric if:
# (header["key1"] == 1 or header["key1"] == 2) and (header["key2"] == "yes" or header["key2"] == "true")
POLARIMETRIC_HEADER_REQUIREMENTS = {
    # range(2, 2**16) is a hack to mean > 1
    "cryo-nirsp": {"CNMODNST": range(2, 2**16), "CNSPINMD": ["Continuous", "Stepped"]},
    "dl-nirsp": {"DLPOLMD": ["Full Stokes"]},
    "visp": {"VSPPOLMD": ["observe_polarimetric"]},
}


def _vbi_is_polarimetric(header: dict[str, Any]) -> False:
    """VBI is *always* non-polarimetric."""
    return False


def _header_match_is_polarimetric(instrument: str, header: dict[str, Any]) -> bool:
    """
    Determine if a header comes from a polarimetric dataset based on matching values with known polarimetric requirements.

    Requirements are defined in POLARIMETRIC_HEADER_REQUIREMENTS and have the form dict[str, dict[str, Iterable and not str]].
    The meaning of this form is {INSTRUMENT: {KEY: [polarimetric, values, ...], ...}, ...}. A header is considered
    polarimetric iff:

    `all(header[key] in POLARIMETRIC_HEADER_REQUIREMENTS[INSTRUMENT][key]
         for key in POLARIMETRIC_HEADER_REQUIREMENTS[INSTRUMENT].keys()])`

    Parameters
    ----------
    instrument
        Instrument name. Must match the controlled list in the SPEC

    header
        The header to check for polarimetrocity
    """
    requirements = POLARIMETRIC_HEADER_REQUIREMENTS[instrument]
    check_list = []
    for key, reqs in requirements.items():
        header_value = header[key]
        if isinstance(header_value, str):
            header_value = header_value.casefold()
            reqs = [v.casefold() for v in reqs]

        check_list.append(header_value in reqs)

    return all(check_list)


def _fallback_is_polarimetric(header: dict[str, Any]) -> False:
    """
    One-stop function for when we can't figure out what to do with a particular instrument.

    Logs a warning and returns False.
    """
    instrument = header.get("INSTRUME", _NOT_A_DKIST_INSTRUMENT).casefold()
    logger.warning(
        f"Checking polarimetric requiredness for instrument {instrument} is not currently supported."
    )
    return False


class PolarimetricRequired(ConditionalRequirement):

    """
    Conditional requirement that makes keys required if the dataset represents polarimetric data.

    spec_field:
      polarimetric_required: bool
    """

    instrument_to_polarimetric_func_mapping = {
        "vbi": _vbi_is_polarimetric,
        "visp": partial(_header_match_is_polarimetric, instrument="visp"),
        "cryo-nirsp": partial(_header_match_is_polarimetric, instrument="cryo-nirsp"),
        "dl-nirsp": partial(_header_match_is_polarimetric, instrument="dl-nirsp"),
    }

    def __init__(self, header: dict[str, Any]):
        super().__init__(header=header)
        instrument = self.header.get("INSTRUME", _NOT_A_DKIST_INSTRUMENT).casefold()

        is_polarimetric_function = self.instrument_to_polarimetric_func_mapping.get(
            instrument, _fallback_is_polarimetric
        )

        try:
            self.is_polarimetric = is_polarimetric_function(header=self.header)
        except Exception as e:
            logger.warning(
                f"Error encountered when checking polarimetric requiredness. Error was:\n{e}"
            )
            self.is_polarimetric = False

    def check_requiredness(self, spec_fields: frozendict[str, Any]) -> bool:
        """Check if the header was from a polarimetric dataset and the key is `polarimetric_required`."""
        return self.is_polarimetric and spec_fields.get("polarimetric_required", False)



from typing import Any, Iterable
from dataclasses import dataclass

from dkist_fits_specifications.utils import frozendict, schema_type_hint


@dataclass
class ExpansionIndex:
    """
    A class for defining a FITS schema expansion.

    index: the string to be substituted, omitting the surrounding '<', '>'
    size: how many (zero-padded) characters to use for the substituted integer
    values: the iterable of integers to be used as substitution for the index

    Example:
        Using the expansion of:
        ExpansionIndex(index="a", size=3, values=range(1, 6))

        on the keyword 'KEY<a>' would produce the expanded set of keys:
        ['KEY001', 'KEY002', 'KEY003', 'KEY004', 'KEY005']

    """

    index: str
    values: Iterable
    size: int = None

    def __post_init__(self):
        if len(str(max(self.values))) > self.size:
            raise ValueError(
                f"The maximum expansion value ({max(self.values)}) does not fit within the prescribed size ({self.size})."
            )

    def _expanded_keys(self, key: str) -> list[str]:
        """Generate schema entries for expanded keys."""
        return [key.replace(f"<{self.index}>", str(i).zfill(self.size)) for i in self.values]

    def generate(self, keys: list[str]) -> list[str]:
        """Generate the new keys to be added."""
        return_keys = []
        for key in keys:
            if f"<{self.index}>" in key:
                return_keys.extend(self._expanded_keys(key=key))
        long_keys = [k for k in return_keys if len(k) > 8]
        if long_keys:
            raise ValueError(
                f"FITS keywords cannot be more than 8 characters in length. {long_keys} are too long."
            )
        return return_keys


def expand_schema(
    schema: schema_type_hint, expansions: list[ExpansionIndex]
) -> dict[str, frozendict[str, Any]]:
    """Perform a schema expansion given a schema and a list of ExpansionIndexes to apply."""
    expanded_schema = dict()
    for fits_keyword_name, spec_fields in schema.items():
        if "<" not in fits_keyword_name:
            expanded_schema.update({fits_keyword_name: spec_fields})
        else:
            expanded_fits_keywords = [fits_keyword_name]
            for expansion in expansions:
                expanded_fits_keywords.extend(expansion.generate(keys=expanded_fits_keywords))
            expanded_schema.update({k: spec_fields for k in expanded_fits_keywords if "<" not in k})
    return expanded_schema



from typing import Any

from dkist_fits_specifications.utils import frozendict
from dkist_fits_specifications.utils.spec_processors.requiredness_base import (
    _NOT_A_DKIST_INSTRUMENT,
    ConditionalRequirement,
)


class InstrumentRequired(ConditionalRequirement):
    """
    Conditional requirement that makes all keys related to the headers "INSTRUME" required.

    spec_field:
      instrument_required: str
    """

    def __init__(self, header: dict[str, Any]):
        super().__init__(header=header)
        self.instrument = self.header.get("INSTRUME", _NOT_A_DKIST_INSTRUMENT).casefold()

    def check_requiredness(self, spec_fields: frozendict[str, Any]) -> bool:
        """Check if the required instrument matches the instrument in the header."""
        # "false" is importantly different than _NOT_A_DKIST_INSTRUMENT
        return spec_fields.get("instrument_required", "false").casefold() == self.instrument
